"Tools for measuring and optimize the execution time"
# pylint: disable=C0303,C0330

__all__ = [
    "benchmark",
]

from timeit import timeit
from .base import sample


class Time(float):
    "Time formatter"
    units = {"nsec": 1e-9, "usec": 1e-6, "msec": 1e-3, "sec": 1.0}

    def __str__(self):
        scales = [(scale, unit) for unit, scale in Time.units.items()]
        scales.sort(reverse=True)
        for scale, unit in scales:
            if self >= scale:
                break

        return f"{self / scale:.3f} {unit}"

    __repr__ = __str__


def default_timer(fnc, number=100):
    return timeit(fnc, number=100) / number


def benchmark(
    tunable,
    *variables,
    timer=default_timer,
    timer_kwargs=None,
    samples=None,
    label="Time",
    **kwargs,
):
    """
    Crosscheck the result of tunable against the reference.
    
    Parameters
    ----------
    comparison: callable (default = numpy.allclose)
        The function to use for comparison. It is called as fnc(reference, value)
        and should return a value from 0 (False) to 1 (True).
    reference: Any
        The reference value. If None, than the default values are used to produce the result.
    variables: list of str
        Set of variables to sample.
    samples: int
        The number of samples to run. If None, all the combinations are sampled.
    timer_kwargs: dict
        Arguments passed to the timer. For default timer:
        - number: (int) number of iterations
    kwargs: dict
        Variables passed to the compute function. See help(tunable.compute)
    """

    return sample(
        tunable,
        *variables,
        callback=lambda fnc: Time(timer(fnc, **(timer_kwargs or {}))),
        callback_calls=True,
        samples=samples,
        label=label,
        **kwargs,
    )



"""
Function for checking the results
"""
# pylint: disable=C0303,C0330

import operator
from numpy import allclose
from .base import sample
from ..finalize import finalize

__all__ = [
    "crosscheck",
]


def crosscheck(
    tunable,
    *variables,
    comparison=allclose,
    samples=None,
    reference=None,
    label="xcheck",
    **kwargs
):
    """
    Crosscheck the result of tunable against the reference.
    
    Parameters
    ----------
    comparison: callable (default = numpy.allclose)
        The function to use for comparison. It is called as fnc(reference, value)
        and should return a value from 0 (False) to 1 (True).
    reference: Any
        The reference value. If None, than the default values are used to produce the result.
    variables: list of str
        Set of variables to sample.
    samples: int
        The number of samples to run. If None, all the combinations are sampled.
    kwargs: dict
        Variables passed to the compute function. See help(tunable.compute)
    """
    if reference is None:
        reference = finalize(tunable).copy().compute(**kwargs)

    return sample(
        tunable,
        *variables,
        callback=lambda res: comparison(reference, res),
        samples=samples,
        label=label,
        **kwargs
    )



"The Tunable class and derived instances"
# pylint: disable=C0303,C0330

__all__ = [
    "compute",
    "Tunable",
    "tunable",
    "Object",
    "function",
    "Function",
]

import operator
import warnings
from inspect import ismethod
from collections import deque
from collections.abc import Iterable
from hashlib import md5
from dill import dumps
from uuid import uuid4
from dataclasses import dataclass
from typing import Any
from varname import varname as _varname

from .graph import Graph, Node, Key


def varname(caller=1, default=None):
    "Wrapper of varname.varname that silences the warning and returns a default value if given."
    warnings.filterwarnings("error" if default is not None else "ignore")
    try:
        return _varname(caller + 1)
    except Warning:
        return default
    finally:
        warnings.filterwarnings("default")


def compute(obj, **kwargs):
    "Compute the value of a tunable object"
    kwargs.setdefault("maxiter", 3)
    if kwargs["maxiter"] <= 0:
        return obj
    if isinstance(obj, Node):
        kwargs.setdefault("graph", Node(obj).graph)
    if isinstance(obj, Key) and not isinstance(obj, Node):
        obj = kwargs["graph"][obj].value
    try:
        obj = obj.__compute__(**kwargs)
        kwargs["maxiter"] -= 1
        obj = compute(obj, **kwargs)
    except AttributeError:
        pass
    return obj


def tunable(obj, deps=None, label=None, uid=None):
    """
    A tunable object.
    
    Parameters
    ----------
    obj: Any
        The object hold as tunable
    label: str
        A label for the object
    uid: Any
        Unique identifier for the object.
    """
    label = label or varname(default="")
    return Object(obj, deps=deps, label=label, uid=uid).tunable()


@dataclass
class Object:
    "A generic object dataclass"
    obj: Any
    deps: Any = None
    label: str = None
    uid: Any = None

    @classmethod
    def extract_deps(cls, deps):
        "Converts a list of deps into Nodes"
        if deps is None:
            return ()
        if not isinstance(deps, Iterable) or isinstance(deps, (Object, Node)):
            deps = (deps,)
        keys = None
        if isinstance(deps, dict):
            keys = deps.keys()
            deps = deps.values()
        deps = tuple(
            dep.tunable()
            if isinstance(dep, Object)
            else Node(dep)
            if isinstance(dep, Node)
            else dep
            for dep in deps
        )
        if keys:
            return tuple(zip(keys, deps))
        return deps

    def __post_init__(self):
        if isinstance(self.obj, Object):
            if self.uid is None:
                self.uid = self.obj.uid
            if self.label is None:
                self.label = self.obj.label
            self.obj = self.obj.obj

        self.deps = Object.extract_deps(self.deps)

        if self.uid is True:
            self.uid = str(uuid4())

        if self.label is None:
            if isinstance(self.obj, Node):
                self.label = Node(self.obj).label
            else:
                try:
                    self.label = self.obj.__name__
                except AttributeError:
                    self.label = repr(self.obj)

    def tunable(self):
        "Returns a tunable object"
        return Tunable(self)

    @property
    def dependencies(self):
        "Returns the list of dependencies for the Object"
        return Node(self.tunable()).dependencies

    def copy(self, **kwargs):
        "Returns a copy of self"
        kwargs.setdefault("deps", self.deps)
        kwargs.setdefault("label", self.label)
        kwargs.setdefault("uid", self.uid)
        return type(self)(self.obj, **kwargs)

    @property
    def key(self):
        "Get the key used by Tunable"
        key = self.label + "-"

        if self.uid:
            key = key + md5(dumps(self.uid)).hexdigest()
        else:
            try:
                parts = tuple(
                    Key(part) if isinstance(part, Key) else part for part in self
                )
                key = key + md5(dumps(parts)).hexdigest()
            except Exception as _:
                print(_)
                self.uid = str(uuid4())
                return self.key

        return Key(key)

    def __iter__(self):
        yield self.obj
        yield from self.deps

    def __compute__(self, **kwargs):
        # pylint: disable=W0108
        cmpt = lambda obj: compute(obj, **kwargs)
        deque(map(cmpt, self.deps))
        return self.obj

    @property
    def __label__(self):
        return self.label

    @property
    def __dot_attrs__(self):
        return dict(shape="rect")


def function(fnc, *args, **kwargs):
    """
    A tunable function call.
    
    Parameters
    ----------
    fnc: callable
        The function to call
    args: tuple
        List of args for the function
    kwargs: dict
        List of named args for the function
    """
    return Function(fnc, args=args, kwargs=kwargs).tunable()


@dataclass
class Function(Object):
    "The Function dataclass"

    args: tuple = None
    kwargs: dict = None

    labels = {
        "call": "(...)",
        "add": "+",
        "sub": "-",
        "mul": "*",
        "div": "/",
        "truediv": "//",
        "eq": "==",
        "ne": "!=",
        "ge": ">=",
        "gt": ">",
        "le": "<=",
        "lt": "<",
    }

    def __post_init__(self):
        if not callable(self.fnc):
            raise TypeError("The first argument of Function must be callable")

        self.args = Object.extract_deps(self.args)
        self.kwargs = dict(Object.extract_deps(self.kwargs))

        super().__post_init__()

    @property
    def fnc(self):
        "Alias of obj"
        return self.obj

    def copy(self, **kwargs):
        "Returns a copy of self"
        kwargs.setdefault("args", self.args)
        kwargs.setdefault("kwargs", self.kwargs)
        return super().copy(**kwargs)

    def __iter__(self):
        yield from super().__iter__()
        yield from self.args
        yield from self.kwargs.values()

    def __call__(self, *args, **kwargs):
        tmp = self.kwargs.copy()
        tmp.update(kwargs)
        return self.copy(args=(self.args + args), kwargs=tmp).tunable()

    def __compute__(self, **kwargs):
        # pylint: disable=W0108
        cmpt = lambda obj: compute(obj, **kwargs)
        fnc = cmpt(super().__compute__(**kwargs))
        args = tuple(map(cmpt, self.args))
        kwargs = dict(zip(self.kwargs.keys(), map(cmpt, self.kwargs.values())))
        res = fnc(*args, **kwargs)
        if res is None:
            if ismethod(fnc) or fnc is setattr:
                return args[0]
            if hasattr(fnc, "__self__"):
                return fnc.__self__
            if args:
                return args[0]
        return res

    @property
    def __label__(self):
        label = self.label
        if label.startswith("__"):
            label = label[2:]
        if label.endswith("__"):
            label = label[:-2]
        if label in Function.labels:
            return Function.labels[label]

        if (
            label in ("getattr", "setattr", "callattr")
            and self.args
            and len(self.args) >= 2
        ):
            return (
                "."
                + self.args[1]
                + (
                    "=..."
                    if label == "setattr"
                    else "(...)"
                    if label == "callattr" and (len(self.args) > 2 or self.kwargs)
                    else "()"
                    if label == "callattr"
                    else ""
                )
            )

        return label

    @property
    def __dot_attrs__(self):
        return dict(shape="oval")


def callattr(self, key, *args, **kwargs):
    "Get attribute from self and then calls it with args, kwargs"
    return getattr(self, key)(*args, **kwargs)


class Tunable(Node, bind=False):
    "A class that turns any operation into a graph node"

    def __init__(self, obj):
        if not isinstance(obj, Object):
            raise TypeError("Expected an Object")

        Node.__init__(Node(self), obj.key, obj.copy())

    def __setattr__(self, key, value):
        tmp = Node(function(setattr, Node(self).copy(), key, value))
        Key(self).key = tmp.key
        Graph(self).backend = tmp.graph.backend

    def __call__(self, *args, **kwargs):
        tmp = Node(self).value
        if (
            isinstance(tmp, Function)
            and tmp.fnc is getattr
            and len(tmp.args) == 2
            and not tmp.kwargs
        ):
            value = Function(callattr, args=tmp.args + args, kwargs=kwargs)
            graph = Graph(self).copy()
            del graph[Key(self)]
            graph[value.key] = value
            return Tunable(graph[value.key])
        return function(Tunable(Node(self).copy()), *args, **kwargs)

    def __repr__(self):
        return "Tunable(%s)" % Node(self).key

    def __bool__(self):
        raise TypeError

    __nonzero__ = __bool__

    def __getstate__(self):
        return Node(self).key

    def __compute__(self, **kwargs):
        # pylint: disable=W0613
        return Node(self).value


def default_operator(fnc):
    "Default operator wrapper"
    return lambda self, *args, **kwargs: function(
        fnc, Node(self).copy(), *args, **kwargs
    )


def add_operators(cls, wrapper):
    "Wraps all the operators with wrapper and adds them to the class"
    setattr(cls, "__getattr__", wrapper(getattr))

    for fnc in dir(operator):
        if fnc.startswith("__"):
            try:
                fnc2 = getattr(operator, fnc[2:-2])
            except AttributeError:
                try:
                    fnc2 = getattr(operator, fnc[2:-1])
                except AttributeError:
                    continue

            assert fnc2 is getattr(operator, fnc)
            setattr(cls, fnc, wrapper(fnc2))

    def roperator(fnc):
        "Default operator wrapper"
        return lambda arg1, arg2: function(fnc, arg2, arg1)

    rops = (
        "__radd__",
        "__rmod__",
        "__rmul__",
        "__rpow__",
        "__rsub__",
        "__rtruediv__",
        "__rfloordiv__",
    )
    for fnc in rops:
        setattr(cls, fnc, roperator(getattr(operator, fnc[3:-2])))


add_operators(Tunable, default_operator)



"A Python tool for optimizing and crosschecking calculations"

from .graph import *
from .tunable import *
from .variable import *
from .finalize import *
from .class_utils import *
from .tools import *



"Tunable variables"

__all__ = [
    "variable",
    "Variable",
    "Permutation",
]

from collections.abc import Iterable
from dataclasses import dataclass
from typing import Any
from itertools import permutations
from math import factorial
from .tunable import Tunable, Object, varname


def variable(var, default=None, label=None, uid=None):
    """
    A tunable variable.

    Parameters
    ----------
    var: Iterable
        An iterator over the possible values of the variable
    default: Any
        The default value for the variable. If None the first element is used.
    label: str
        A label used to identify the variable
    uuid: Any
        Unique identifier for the variable.
    """
    label = label or varname()
    return Variable(var, default=default, label=label, uid=uid).tunable()


class Value:
    "Simple class to hold the value of the variable"
    __slots__ = ["value"]

    @property
    def fixed(self):
        "Returns if the value has been fixed"
        return hasattr(self, "value")


@dataclass
class Variable(Object):
    "The Variable dataclass"
    default: Any = None
    _value: Value = None

    def __post_init__(self):
        if not isinstance(self.var, Iterable):
            raise TypeError("The first argument of Variable must be iterable")

        if not hasattr(self.var, "__len__"):
            self.obj = tuple(self.obj)

        if self.default is None:
            try:
                self.default = next(iter(self.values))
            except StopIteration:
                raise ValueError("Given an empty range for the variable")

        if self.default not in self.values:
            raise ValueError("Default value not in variable's value")

        if self._value is None:
            self._value = Value()

        if self.size < 2 and not self.fixed:
            self.value = self.default

        assert isinstance(self._value, Value), "Value must be of Value type"

        if self.uid is None:
            self.uid = True

        super().__post_init__()

    @property
    def fixed(self):
        "Returns if the value has been fixed"
        return self._value.fixed

    def fix(self, value=None):
        "Fixes the value of the variable"
        if self.fixed and value != self.value:
            raise RuntimeError("Cannot change a value that has been fixed")
        if value is None:
            value = self.default
        if isinstance(value, Variable):
            value = value.tunable()
        if isinstance(value, Iterable) and not isinstance(value, str):
            value = tuple(value)
        if not isinstance(value, Tunable) and value not in self.values:
            raise ValueError("Value %s not compatible with variable" % (value,))
        self._value.value = value

    @property
    def value(self):
        "Returns the value of the variable. If it is not fixed, the default value is returned."
        if self.fixed:
            return self._value.value
        return self.tunable()

    @value.setter
    def value(self, value):
        self.fix(value)

    @property
    def var(self):
        "Alias of obj"
        return self.obj

    def __iter__(self):
        yield self.var
        if self.fixed:
            yield self.value

    def __eq__(self, other):
        if self.fixed:
            if isinstance(other, Variable):
                if not other.fixed:
                    return False
                other = other.value
            return self.value == other
        if isinstance(other, Variable):
            return super().__eq__(other)
        return False

    @property
    def size(self):
        "Returns the size of the variable range"
        return len(self.var)

    @property
    def values(self):
        "Returns the value in the variable range"
        return self.var

    def __compute__(self, **kwargs):
        if not self.fixed:
            self.value = self.default
        return self.value

    @property
    def __dot_attrs__(self):
        return dict(shape="diamond", color="green" if self.fixed else "red")

    def copy(self, reset=False, reset_value=False, **kwargs):
        "Returns a copy of self"
        kwargs.setdefault("default", self.default)
        if reset or reset_value:
            kwargs.setdefault("_value", None)
        else:
            kwargs.setdefault("_value", self._value)
        if reset:
            kwargs.setdefault("uid", True)
        else:
            kwargs.setdefault("uid", self.uid)
        return super().copy(**kwargs)

    def __repr__(self):
        return "%s(%s%s)" % (
            type(self).__name__,
            self.var,
            (", value=%s" % str(self.value)) if self.fixed else "",
        )


class Permutation(Variable):
    "Permutations of the given list"

    @property
    def size(self):
        return factorial(len(self.var))

    @property
    def values(self):
        return permutations(self.var)



from tuneit.graph import visualize
from tuneit.tunable import *
from tuneit.variable import *
from tuneit.tunable import Tunable
from tuneit.finalize import finalize
from pytest import raises


def test_finalize():
    with raises(TypeError):
        finalize(1)

    a = variable(range(10), default=2)

    assert finalize(a)[finalize(a).value] == finalize(a).value

    c = variable(range(10))
    b = finalize(a * a + c)

    assert set(b.variables) == set([finalize(a).key, finalize(c).key])
    assert b.tunable_variables == b.variables
    assert b.compute() == 4
    assert b.fixed_variables == b.variables
    assert not b.tunable_variables

    assert len(b.functions) == 2
    assert not b.depends_on(1)
    assert b.depends_on(a)
    assert b.depends_on(finalize(a).value)

    b = b.copy(reset=True)
    assert b.tunable_variables == b.variables
    assert finalize(a).value.fixed

    d = b.copy()
    assert d.compute() == 4
    assert b.tunable_variables == b.variables
    assert d.fixed_variables == b.variables

    b.fix("a")
    b.fix(finalize(c).value, 1)
    assert b.compute() == 5
    assert b.fixed_variables == b.variables

    with raises(KeyError):
        b.fix("foo")

    a = variable(range(10), uid=True)
    with raises(KeyError):
        finalize(a * b).fix("a")



"""
Utils for creating tunable classes
"""
# pylint: disable=C0103

__all__ = [
    "TunableClass",
    "tunable_property",
    "derived_property",
    "derived_method",
    "alternatives",
]

from functools import partial, wraps
from .graph import Graph, Node, visualize
from .tunable import Tunable, tunable, Function, function
from .variable import Variable, variable
from .finalize import finalize


class TunableClass:
    "Base class for tunable classes"

    def __init__(self, value=None):
        self.value = value

    @property
    def value(self):
        "Returns the underlying tunable value of the class"
        return self._value

    @value.setter
    def value(self, value):
        if isinstance(value, Node):
            self._value = Tunable(value)
        elif isinstance(value, TunableClass):
            self._value = value.value
        else:
            self._value = tunable(value, label="value")

    @property
    def node(self):
        "Returns the node of the tunable value"
        return finalize(self.value)

    @property
    def graph(self):
        "Returns the underlying graph"
        return Graph(self.value)

    @property
    def __graph__(self):
        return self.value

    def visualize(self, **kwargs):
        "Visualizes the class graph"
        return visualize(self.value, **kwargs)

    def compute(self, **kwargs):
        "Computes the class graph"
        self.value = tunable(self.node.compute(**kwargs), label="value")

    @property
    def result(self):
        "Returns the value of the class after computing"
        self.compute()
        return self.node.value.obj

    @property
    def variables(self):
        "List of variables in the class graph"
        return self.node.variables

    @property
    def tunable_variables(self):
        "List of tunable variables in the class graph"
        return self.node.tunable_variables

    @property
    def fixed_variables(self):
        "List of fixed variables in the class graph"
        return self.node.fixed_variables


class tunable_property(property):
    """
    Returns a tunable property of the class.
    The output of a tunable property is threated as a Variable
    """

    @property
    def name(self):
        "Name of the property"
        return self.fget.__name__

    @property
    def key(self):
        "Key where to store the property value"
        return getattr(self, "_key", "_" + self.name)

    @key.setter
    def key(self, value):
        self._key = value

    def __get__(self, obj, owner):
        try:
            return getattr(obj, self.key)
        except AttributeError:
            var = super().__get__(obj, owner)
            if isinstance(var, Variable):
                var.label = self.name
            else:
                var = Variable(super().__get__(obj, owner), label=self.name)
            setattr(obj, self.key, var)
            return self.__get__(obj, owner)

    def __set__(self, obj, value):

        var = self.__get__(obj, type(obj))

        if isinstance(value, Variable):
            setattr(obj, self.key, value)
        else:
            var.value = value


def skip_n_args(fnc, num):
    "Decorator that calls a function skipping the first n arguments"

    @wraps(fnc)
    def wrapped(*args, **kwargs):
        return fnc(*args[num:], **kwargs)

    return wrapped


def derived_method(*deps):
    """
    Returns a method that depends on a tunable property.
    """

    assert all(
        (isinstance(dep, tunable_property) for dep in deps)
    ), "TODO: improve check"

    def decorator(fnc):
        @wraps(fnc)
        def derived(self, *args, **kwargs):
            _deps = (dep.__get__(self, type(self)) for dep in deps)
            _deps = tuple(dep.value for dep in _deps)
            if any((isinstance(dep, Tunable) for dep in _deps)):
                return Function(
                    fnc, deps=_deps, args=(self,) + args, kwargs=kwargs
                ).tunable()
            return fnc(self, *args, **kwargs)

        return derived

    return decorator


class derived_property(property):
    """
    Returns a property that depends on a tunable property.
    """

    @property
    def name(self):
        "Name of the property"
        return self.fget.__name__

    def __new__(cls, *deps, **kwargs):
        if "deps" not in kwargs:
            return partial(derived_property, deps=deps)
        return super().__new__(cls, *deps, **kwargs)

    def __init__(self, *args, deps=None, **kwargs):
        super().__init__(*args, **kwargs)

        self.deps = deps or ()
        self.__name__ = self.name

        assert all(
            (isinstance(dep, tunable_property) for dep in self.deps)
        ), "TODO: improve check"

    def __get__(self, obj, owner):
        if obj is None:
            return self.fget
        deps = (dep.__get__(obj, owner) for dep in self.deps)
        deps = tuple(dep.value for dep in deps)
        if any((isinstance(dep, Tunable) for dep in deps)):
            return Function(
                self.fget, deps=deps, args=(obj,), label=self.name
            ).tunable()
        return super().__get__(obj, owner)


class alternatives(dict):
    """
    Class for defining alternative implementations of a function

    Usage
    -----
    @alternatives(fnc1, label = fnc2, other_label = fnc3)
    def fnc(*args, **kwargs):
        ...
    """

    @classmethod
    def args_to_kwargs(cls, *args):
        "Turns args into kwargs using as a key the arg name"

        kwargs = {}
        for arg in args:
            if isinstance(arg, dict):
                kwargs.update(arg)
            elif hasattr(arg, "__name__"):
                kwargs[arg.__name__] = arg
            else:
                raise ValueError(f"{arg} has no __name__")

        return kwargs

    def __init__(self, *args, **kwargs):
        super().__init__(**self.args_to_kwargs(*args), **kwargs)
        self.default = next(iter(self))

    @wraps(dict.update)
    def update(self, *args, **kwargs):
        super().update(**self.args_to_kwargs(*args), **kwargs)

    @property
    def default(self):
        "The default value"
        return self._default

    @default.setter
    def default(self, key):
        if key not in self:
            raise KeyError(f"{key} unknown alternative")
        self._default = key
        wraps(self[key])(self)
        self.__name__ = key

    def add(self, fnc):
        "Adds a value to the alternatives"
        kwargs = self.args_to_kwargs(fnc)
        self.update(kwargs)
        return next(iter(kwargs))

    def __call__(self, *args, _key=None, **kwargs):
        if len(args) == 1 and callable(args[0]):
            self.default = self.add(args[0])
            return self

        if _key:
            return self[_key](*args, **kwargs)

        return function(
            self,
            *args,
            _key=variable(self.keys(), default=self.default, label=self.__name__),
            **kwargs,
        )



from tuneit.graph import visualize
from tuneit.tunable import *
from tuneit.variable import *
from tuneit.tunable import Tunable
from tuneit.finalize import finalize
from pytest import raises


def test_variable():
    A = Variable(range(10), default=2)
    a = A.tunable()
    C = Variable(range(10))
    c = C.tunable()
    b = a * a + c * tunable(1)

    assert not A.fixed
    assert not C.fixed
    assert compute(b) == 4
    assert A.fixed
    assert C.fixed
    assert A.value == 2
    assert C.value == 0

    assert visualize(b).source

    with raises(TypeError):
        variable(1)

    with raises(ValueError):
        variable(range(10), default=11)

    d = Variable(range(10))
    assert d == Variable(range(10), uid=d.uid)
    assert d != Variable(range(10))
    assert d.size == 10
    with raises(ValueError):
        d.value = 11

    d.value = 2
    assert d == 2
    assert "2" in repr(d)
    with raises(RuntimeError):
        d.value = 3

    d = Variable(range(10))
    d.value = Variable(range(10))

    d = Permutation((1, 2, 3))
    d.value = (3, 2, 1)
    assert d.size == 6

    d = Variable([0])
    assert d.fixed
    assert d.value == 0

    with raises(ValueError):
        d = Variable([])

    d = Variable((i for i in [1, 2, 3]))
    assert d.values == (1, 2, 3)



"""
Graph and node structures.
"""

__all__ = [
    "visualize",
]

from collections import deque
from collections.abc import Iterable
from .meta import CastableType


class Graph(metaclass=CastableType, attrs=["backend"]):
    """
    A Graph class.
    """

    def __init__(self, graph=None):
        if isinstance(graph, Graph):
            graph = graph.backend
        self.backend = {} if graph is None else dict(graph)

    def __getitem__(self, key):
        if isinstance(key, Key):
            key = Key(key).key
        node = Node(self)
        Key(node).key = key
        return node

    def __setitem__(self, key, value):
        if isinstance(key, Key):
            key = Key(key).key
        self.backend[key] = value

    def __delitem__(self, key):
        if isinstance(key, Key):
            key = Key(key).key
        del self.backend[key]

    def __getattr__(self, key):
        return getattr(self.backend, key)

    def __contains__(self, key):
        if isinstance(key, Key):
            key = Key(key).key
        return key in self.backend

    def __eq__(self, value):
        if isinstance(value, Graph):
            return self.backend == Graph(value).backend
        return self.backend == value

    def __repr__(self):
        return "Graph(%s)" % self

    def __str__(self):
        return str(self.backend)

    def __iter__(self):
        return iter(self.backend)

    def visualize(self, **kwargs):
        """
        Visualizes the complete graph.
        For more details see help(visualize).
        """
        return visualize(self, **kwargs)

    def update(self, value):
        "Updates the content of the dictionary"
        if isinstance(value, Graph):
            value = Graph(value).backend
        return self.backend.update(value)

    def copy(self):
        "Shallow copy of a Graph"
        return Graph(self.backend.copy())


class Key(metaclass=CastableType, attrs=["key"]):
    "Namespace for the keys of tunable objects"

    def __init__(self, key):
        if isinstance(key, Key):
            key = Key(key).key
        self.key = key

    def __repr__(self):
        return "Key(%s)" % repr(self.key)

    def __str__(self):
        return str(self.key)

    def __eq__(self, value):
        if isinstance(value, Key):
            return self.key == Key(value).key
        return self.key == value

    def __getattr__(self, key):
        return getattr(self.key, key)

    def __hash__(self):
        return hash(self.key)


class Node(Graph, Key, bind=False):
    """
    A node of the graph
    """

    def __init__(self, key, value=None):
        Graph.__init__(self.graph, join_graphs(value))
        Key.__init__(self.key, key)

        if key not in self.graph:
            self.graph[key] = value
        elif not isinstance(value, Graph) or self == value:
            raise KeyError("%s already in graph and the value will not be changed")

        for part in self:
            try:
                Key.__cast__(part)
            except TypeError:
                pass

    @property
    def key(self):
        "Returns the key of the node"
        return Key(self)

    @property
    def label(self):
        "Return the label used in the dot graph"
        try:
            label = self.value.__label__
            assert isinstance(label, str)
            return label
        except (AttributeError, AssertionError):
            return self.key.split("-")[0]

    @property
    def dot_attrs(self):
        "Return the node attributes used in the dot graph"
        try:
            attrs = self.value.__dot_attrs__
            assert isinstance(attrs, dict)
            return attrs
        except (AttributeError, AssertionError):
            return {}

    @property
    def value(self):
        "Value of the node"
        return self.graph.backend[Key(self).key]

    @value.setter
    def value(self, value):
        "Sets the value of the node"
        self.__init__(self.key, value)

    @property
    def graph(self):
        "Returns the graph that the node is part of"
        return Graph(self)

    def __eq__(self, value):
        if isinstance(value, Node):
            return self.key == value and self.value == Node(value).value
        if isinstance(value, Key):
            return self.key == value
        return self.value == value

    def __iter__(self):
        try:
            yield from self.value
        except TypeError:
            yield self.value

    def __repr__(self):
        return "Node(%s)" % (self.key)

    def __str__(self):
        return str(self.value)

    def __copy__(self):
        return Node(Key(self), Graph(self))

    def copy(self):
        return self.__copy__()

    @property
    def first_dependencies(self):
        "Iterates over the dependencies"
        for val in self:
            if isinstance(val, Key):
                yield Key(val)

    @property
    def dependencies(self):
        "Iterates over the dependencies"
        deps = [self.key]
        yield deps[0]
        for val in self:
            if isinstance(val, Key):
                val = self.graph[val]
            if isinstance(val, Node):
                for dep in Node(val).dependencies:
                    if dep not in deps:
                        deps.append(dep)
                        yield dep

    def visualize(self, **kwargs):
        """
        Visualizes the graph up to this node.
        For more details see help(visualize).
        """
        return visualize(self, **kwargs)


def join_graphs(graphs):
    "Joins all the graphs in graphs into a unique instance"

    if isinstance(graphs, Graph):
        return Graph(graphs).copy()

    if isinstance(graphs, dict):
        return join_graphs(graphs.values())

    if isinstance(graphs, str):
        return Graph()

    if isinstance(graphs, Iterable):
        graph = Graph()
        graphs = filter(lambda obj: isinstance(obj, Graph), graphs)
        deque(map(graph.update, graphs))
        return graph

    return Graph()


def visualize(graph, start=None, end=None, **kwargs):
    "Visualizes the graph returning a dot graph"
    assert isinstance(graph, Graph), "graph must be of type Graph"

    if isinstance(graph, Node):
        end = end or Node(graph).key

    graph = Graph(graph)

    if end is not None:
        if end not in graph:
            raise KeyError("Given end %s not in graph" % end)
        end = graph[end]
        keys = tuple(end.dependencies)
    else:
        keys = tuple(graph.keys())

    if isinstance(start, Node):
        start = Node(start).key

    if start and start not in keys:
        raise KeyError("Given start %s not in graph" % end)

    dot = default_graph(**kwargs)

    for key in keys:
        node = graph[key]

        if start and start not in node.dependencies:
            continue

        dot.node(str(key), node.label, **node.dot_attrs)

        for dep in node.first_dependencies:
            if start and start not in graph[dep].dependencies:
                continue
            dot.edge(str(dep), str(key))

    return dot


def default_graph(**kwargs):
    "Defines the default options for kwargs"
    try:
        # pylint: disable=import-outside-toplevel
        from graphviz import Digraph
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            """
            Visualize needs graphviz to be installed.
            Please run `pip install --user graphviz`.
            """
        )

    kwargs.setdefault("graph_attr", {})
    kwargs["graph_attr"].setdefault("rankdir", "LR")
    return Digraph(**kwargs)



from tuneit.graph import visualize, Node
from tuneit.tunable import *
from tuneit.tunable import Tunable
from pytest import raises


class unpickable:
    def __getstate__(self):
        raise NotImplementedError


def test_object():
    zero = Object(0)
    assert zero.uid == None
    assert zero.key == Object(0).key
    assert zero == Object(zero)
    assert Object(0, uid=True).key != Object(0, uid=True).key

    assert compute(zero) == 0
    assert (
        Object(0, label="foo", uid="bar").key == Object(1, label="foo", uid="bar").key
    )

    a = Object(unpickable())
    assert a.key == a.key
    assert a.key != Object(unpickable()).key

    one = Object(1, deps=zero)
    assert zero.key in one.dependencies


def test_function():
    with raises(TypeError):
        function(1)

    zero = Object(0)
    assert compute(Function(str, args=(zero))) == "0"
    assert Function(int.__str__).__label__ == "str"
    assert Function(int.__mul__).__label__ == "*"
    assert Function(int.__setattr__).__label__ == "setattr"
    assert Function(int.__setattr__, args=(0, "foo")).__label__.startswith(".foo=")

    one = Object(1)
    fnc = Function(str, args=zero, kwargs={"one": one})
    assert zero.key in fnc.dependencies
    assert one.key in fnc.dependencies

    lst = Object([1, 2]).tunable()
    lst = lst.append(3)
    assert compute(lst) == [1, 2, 3]

    lst = [1, 2]
    lst = function(lst.append, 3)
    assert compute(lst) == [1, 2, 3]


def one():
    return tunable(tunable(1, uid=True))


def test_tunable():
    with raises(TypeError):
        Tunable(1)

    a = tunable(2)
    b = a * a * one()

    assert compute(b) == 4

    assert Node(b).copy() == b

    to_string = tunable(lambda *args: str(*args))
    c = to_string(b)
    assert compute(c) == "4"

    to_string.__name__ = "foo"
    assert compute(to_string).__name__ == "foo"

    to_string = Function(str)
    c = to_string(b)
    assert compute(c) == "4"

    assert repr(b).startswith("Tunable")
    assert visualize(b).source

    with raises(TypeError):
        bool(tunable(1))



"""
High Level vision of Tunable object
"""

__all__ = [
    "finalize",
]

from .graph import Node, Key
from .variable import Variable
from .tunable import Object, Function, compute


def finalize(tunable):
    "Returns a finalized tunable object that has several high-level functions"
    if not isinstance(tunable, Node):
        raise TypeError("Only tunable objects can be finalized")
    return HighLevel(tunable)


class HighLevel(Node):
    "HighLevel view of a Node"

    @property
    def variables(self):
        "List of dependencies that are a variable"
        return tuple(
            str(dep) for dep in self.dependencies if isinstance(self[dep], Variable)
        )

    @property
    def functions(self):
        "List of dependencies that are a functions"
        return tuple(
            dep for dep in self.dependencies if isinstance(self[dep], Function)
        )

    @property
    def tunable_variables(self):
        "List of variables that are tunable"
        return tuple(var for var in self.variables if not self[var].fixed)

    @property
    def fixed_variables(self):
        "List of variables that are fixed"
        return tuple(var for var in self.variables if self[var].fixed)

    def depends_on(self, value):
        "Returns true if the given value is in the graph"
        if isinstance(value, Key):
            return Key(value).key in self.dependencies
        if isinstance(value, Object):
            return self.depends_on(value.key)
        return False

    def __getitem__(self, key):
        if isinstance(key, Object):
            key = key.key
        return self.graph[key].value

    def __setitem__(self, key, value):
        self.graph[key] = value

    def __copy__(self):
        return HighLevel(super().__copy__())

    def copy(self, reset=False, reset_tunable=True):
        "Copy the content of the graph unrelating the tunable variables"
        res = self.__copy__()

        if reset:
            for var in res.variables:
                res[var] = res[var].copy(reset_value=True)
        elif reset_tunable:
            for var in res.tunable_variables:
                res[var] = res[var].copy(reset_value=True)

        return res

    def get_variable(self, variable):
        "Returns the varible corresponding to var"
        if isinstance(variable, Variable):
            variable = variable.key
        if isinstance(variable, Key):
            variable = Key(variable).key

        if not variable in self.variables:
            # Smart search
            matches = list(
                filter(lambda var: var.startswith(variable + "-"), self.variables)
            )
            if len(matches) > 1:
                raise KeyError(
                    "More than one variable matched to %s: %s" % (variable, matches)
                )
            if len(matches) == 0:
                raise KeyError("%s is not a variable of %s" % (variable, self))
            variable = matches[0]

        return self[variable]

    def fix(self, variable, value=None):
        "Fixes the value of the variable"
        self.get_variable(variable).fix(value)

    def compute(self, **kwargs):
        "Computes the result of the Node"
        kwargs.setdefault("graph", self.graph)
        return compute(self.value, **kwargs)



"""
Base methods for the tunable tools
"""
# pylint: disable=C0303,C0330

__all__ = [
    "sample",
]

import operator
import random
from functools import reduce
from itertools import product
from tabulate import tabulate
from ..finalize import finalize


class Sampler:
    "Base class for sampling values of a tunable object"

    def __init__(
        self,
        tunable,
        variables=None,
        n_samples=None,
        callback=None,
        callback_calls=False,
        label=None,
        **kwargs,
    ):
        "Initializes the tunable object and the variables"

        self.tunable = finalize(tunable).copy()
        self.compute_kwargs = kwargs

        if callback:
            self.callback = callback
        self.callback_calls = callback_calls

        if label:
            self.label = label

        if variables:
            self.variables = tuple(
                str(tunable.get_variable(var).key) for var in variables
            )

            set_vars = set(self.variables)
            set_tunable = set(self.tunable.tunable_variables)
            if not set_vars <= set_tunable:
                raise ValueError(
                    f"Variable(s) {set_vars-set_tunable} have been fixed and cannot be tuned"
                )

            # Fixes the tunable variable not involved
            for var in set_tunable.difference(set_vars):
                self.tunable.fix(var)

        else:
            self.variables = self.tunable.tunable_variables

        if n_samples:
            self.n_samples = n_samples

    @property
    def max_samples(self):
        "Size of the parameter space (product of variables' size)"
        lens = tuple(self.tunable[var].size for var in self.variables)
        return reduce(operator.mul, lens)

    @property
    def n_samples(self):
        "Number of samples"
        return getattr(self, "_n_samples", self.max_samples)

    @n_samples.setter
    def n_samples(self, value):
        if not (isinstance(value, int) and value > 0):
            raise ValueError("n_samples must be a positive integer")
        self._n_samples = min(value, self.max_samples)

    def __len__(self):
        self.n_samples

    @property
    def samples(self):
        "Samples of the parameters space"

        iters = tuple(self.tunable[var].values for var in self.variables)
        values = product(*iters)

        if self.n_samples >= self.max_samples:
            return tuple(values)

        idxs = sorted(random.sample(range(self.max_samples), self.n_samples))

        return tuple(
            val for idx, val in filter(lambda _: _[0] in idxs, enumerate(values))
        )

    def sample_values(self):
        "Returns the sampled values"
        return list(self)

    @property
    def callback(self):
        return getattr(self, "_callback", lambda _: _)

    @callback.setter
    def callback(self, value):
        "Function to be called on the result"
        if not callable(value):
            raise TypeError("callback must be a callable")
        self._callback = value

    def __iter__(self):
        for params in self.samples:
            tmp = self.tunable.copy()
            for var, val in zip(self.variables, params):
                tmp.fix(var, val)
            try:
                if self.callback_calls:
                    result = self.callback(lambda: tmp.compute(**self.compute_kwargs))
                else:
                    result = self.callback(tmp.compute(**self.compute_kwargs))
            except Exception as err:
                result = err
            yield params, result

    @property
    def label(self):
        "Label used for the result"
        return getattr(self, "_label", self.tunable.label)

    @label.setter
    def label(self, value):
        self._label = str(value)

    @property
    def headers(self):
        "Headers for the values returned by the sampler"
        return tuple(self.tunable[var].label for var in self.variables) + (self.label,)

    def tabulate(self, **kwargs):
        "Returns a table of the values"
        kwargs.setdefault("headers", self.headers)
        return tabulate((params + (repr(result),) for params, result in self), **kwargs)

    def _repr_html_(self):
        return self.tabulate(tablefmt="html")


def sample(tunable, *variables, samples=100, **kwargs):
    """
    Samples the value of the tunable object
    
    Parameters
    ----------
    variables: list of str
        Set of variables to sample.
    samples: int
        The number of samples to run. If None, all the combinations are sampled.
    kwargs: dict
        Variables passed to the compute function. See help(tunable.compute)
    """
    return Sampler(tunable, variables=variables, n_samples=samples, **kwargs)



"Defines the CastableType metaclass used by the module"
# pylint: disable=C0303,C0330

__all__ = [
    "CastableType",
]


class Slot:
    """
    An holder of the slot value. 
    Needed for passing by reference the slots between views of the class.
    """

    __slots__ = ["value"]

    @classmethod
    def getter(cls, key):
        "Returns the getter function"

        def fget(self):
            return getattr(type(self), "__values__").__get__(self)[key].value

        return fget

    @classmethod
    def setter(cls, key):
        "Returns the setter function"

        def fset(self, value):
            getattr(type(self), "__values__").__get__(self)[key].value = value

        return fset


class CastableType(type):
    """
    A metaclass for castable classes.

    >>> from tunable.meta import CastableType
    
    >>> class Foo(metaclass=CastableType, attrs=["foo"]):
    ...     pass
    
    >>> class Bar(Foo, attrs=["bar"], bind=False):
    ...     pass
    
    >>> Bar.__attrs__
        ["Foo_foo", "Bar_bar"]
    >>> issubclass(Bar, Foo)
        True
    >>> bar = Bar()
    >>> hasattr(bar, "foo")
        False

    **Note**: since we initialized Bar with bind=False
    it has not inherit the methods from the parent class.
    
    To access the methods of the parent class one
    needs to cast the instance to the parent class.
    
    >>> foo = Foo(bar)
    >>> hasattr(foo, "foo")
        True    
    """

    @classmethod
    def caster(cls, scls):
        "Returns the cast function for the given subclass (scls)"

        def __cast__(self):
            if not isinstance(self, scls):
                raise TypeError("Cannot cast %s" % repr(self))
            for key in tuple(getattr(type(self), "__values__").__get__(self).keys()):
                if key not in scls.__attrs__:
                    del getattr(type(self), "__values__").__get__(self)[key]
            object.__setattr__(self, "__class__", scls)

        return __cast__

    @classmethod
    def __prepare__(cls, name, bases, **kwargs):
        "Collects the attrs from bases"

        attrs = set()
        class_attrs = {}
        for base in bases:
            if isinstance(base, CastableType):
                attrs.update(base.__attrs__)

        for attr in kwargs.get("attrs", ()):
            key = "%s_%s" % (name, attr)
            class_attrs[attr] = property(Slot.getter(key), Slot.setter(key))
            attrs.add(key)

        class_attrs["__attrs__"] = tuple(attrs)
        class_attrs["__slots__"] = ("__values__",)
        return class_attrs

    def __new__(cls, name, bases, class_attrs, **kwargs):
        "Checks that all attrs of bases are subset of __attrs__"

        attrs = set(class_attrs["__attrs__"])
        bases = list(bases)
        bind = kwargs.get("bind", True)
        for base in list(bases):
            if isinstance(base, CastableType):
                assert attrs.issuperset(
                    base.__attrs__
                ), """
                Given attrs do not match with base class
                """
                if not bind:
                    bases.remove(base)

        self = super().__new__(cls, name, tuple(bases), class_attrs)
        self.__cast__ = cls.caster(self)
        return self

    def __call__(cls, *args, **kwargs):
        "Either calls the class initialization or simply casts"

        args = tuple(args)

        # pylint: disable=E1120
        obj = cls.__new__(cls)
        # pylint: enable=E1120
        values = dict()
        getattr(cls, "__values__").__set__(obj, values)

        if len(args) == 1 and (
            isinstance(args[0], cls) or issubclass(cls, type(args[0]))
        ):
            for attr in obj.__attrs__:
                values[attr] = (
                    getattr(type(args[0]), "__values__")
                    .__get__(args[0])
                    .get(attr, Slot())
                )
        else:
            for attr in obj.__attrs__:
                values[attr] = Slot()
            try:
                obj.__init__(*args, **kwargs)
            except AttributeError:
                pass

        return obj

    def __subclasscheck__(cls, child):
        "Checks if child is subclass of class"
        return isinstance(child, CastableType) and set(child.__attrs__).issuperset(
            cls.__attrs__
        )

    def __instancecheck__(cls, instance):
        "Checks if instance is instance of cls"
        return issubclass(type(instance), cls)



from setuptools import find_packages, setup
from functools import reduce

requirements = [
    "dill",
    "dataclasses",
    "python-varname",
    "tabulate",
    "numpy",
]

extras = {"graph": ["graphviz",]}

extras["all"] = list(set(reduce(lambda a, b: a + b, extras.values())))

classifiers = [
    "Development Status :: 2 - Pre-Alpha",
    "Intended Audience :: Developers",
    "Intended Audience :: Education",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: BSD License",
    "Natural Language :: English",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: 3.6",
]

setup(
    name="tuneit",
    author="Simone Bacchio",
    author_email="s.bacchio@gmail.com",
    url="https://tuneit.readthedocs.io/en/latest",
    download_url="https://github.com/sbacchio/tuneit",
    version="0.0.3",
    license='BSD',
    packages=find_packages(),
    install_requires=requirements,
    extras_require=extras,
    python_requires='>=3',
    description="Tune, benchmark and crosscheck calculations contructing a computational graph",
    long_description=str(open("README.md").read()),
    long_description_content_type="text/markdown",
    classifiers=classifiers,
)



import pytest
from tuneit.graph import Graph, Node, Key


class String(str):
    def __init__(self, val):
        self.__name__ = val
        self.__label__ = val
        self.__dot_attrs__ = {}

    def __iter__(self):
        raise TypeError


def test_graph():
    a = Graph()
    a["foo"] = String("bar")
    b = Node("foo2")
    b.value = (a["foo"], "bar", {"one": 1})

    with pytest.raises(KeyError):
        Node("foo", (a["foo"], "bar"))

    assert isinstance(a["foo"], Node)
    assert "foo" in a
    assert "foo" in list(a)
    assert isinstance(a["foo"].value, str) and a["foo"].key == "foo"
    assert isinstance(a["foo"].value, str) and a["foo"].value == "bar"

    assert a["foo"] == Node("foo", a)
    assert a["foo"] == "bar"
    assert a["foo"] == Key(a["foo"])
    assert str(a["foo"]) == "bar"
    assert a["foo"] != a
    assert a[a["foo"]] == a["foo"]
    assert a["foo"] in a

    assert repr(a).startswith("Graph")
    assert repr(a["foo"]).startswith("Node")
    assert repr(Key(a["foo"])).startswith("Key")

    assert a == a.backend

    assert a.visualize()
    with pytest.raises(KeyError):
        a.visualize(end="bar")
    with pytest.raises(KeyError):
        a.visualize(start="bar")

    assert b.graph.visualize(end=b).source == b.visualize().source
    assert b.visualize().source == b.visualize(start="foo").source
    assert b.visualize().source != b.visualize(start=a["foo2"]).source


def test_dict_methods():
    a = Graph(dict(one=1, two=2, three=3))
    assert a.get("one", None) == 1
    assert a.get("four", None) == None

    b = a.copy()
    assert a == b
    assert b.pop("one") == 1
    assert "one" not in b
    assert "one" in a

    b.update(a)
    assert a == b
